self.__BUILD_MANIFEST=function(s,e){return{__rewrites:{afterFiles:[],beforeFiles:[],fallback:[]},"/":[s,"static/chunks/499-b75ab6c214c5007c.js",e,"static/css/6791d77cb3ee45af.css","static/chunks/pages/index-9792a42cfda807c2.js"],"/_error":["static/chunks/pages/_error-7a92967bea80186d.js"],"/benchmarks":[s,e,"static/css/0096c84fa71d6718.css","static/chunks/pages/benchmarks-29708a196e6ca254.js"],sortedPages:["/","/_app","/_error","/benchmarks"]}}("static/chunks/759-1256f4f6406e6260.js","static/chunks/472-b60185a08ceee629.js"),self.__BUILD_MANIFEST_CB&&self.__BUILD_MANIFEST_CB();